package org.colllib.datastruct;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.colllib.builder.ListBuilder;
import org.colllib.filter.Filter;
import org.colllib.generator.Generator;
import org.colllib.transformer.Transformer;
import org.colllib.transformer.TransformerCollection;
import org.colllib.util.CollectionUtil;

/**
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 * 
 * @author marcel.billen
 */
public class ResultSetWrapper implements Externalizable {
   public List<Pair<String, Class<?>>> columnNamesTypes;
   public List<List<Object>> dataRows;
   private int currentRow;
   
   public static enum Join implements Filter<Pair<ArrayList<List<Object>>, ArrayList<List<Object>>>> {
      OUTER {
         @Override
         public boolean matches(Pair<ArrayList<List<Object>>, ArrayList<List<Object>>> row) {
            return true;
         }
      },
      INNER {
         @Override
         public boolean matches(Pair<ArrayList<List<Object>>, ArrayList<List<Object>>> row) {
            return row.x != null && row.y != null;
         }
      },
      LEFT {
         @Override
         public boolean matches(Pair<ArrayList<List<Object>>, ArrayList<List<Object>>> row) {
            return row.x != null;
         }
      },
      RIGHT {
         @Override
         public boolean matches(Pair<ArrayList<List<Object>>, ArrayList<List<Object>>> row) {
            return row.y != null;
         }
      }
   }
   
   public ResultSetWrapper() {
      this(new ArrayList<Pair<String, Class<?>>>(), new ArrayList<List<Object>>());
   }
   
   public ResultSetWrapper(List<Pair<String, Class<?>>> columnNamesTypes, List<List<Object>> dataRows) {
      this.columnNamesTypes = columnNamesTypes;
      this.dataRows = dataRows;
      this.currentRow = -1;
   }

   public ResultSetWrapper(ResultSet rs) {
      if(rs != null) 
         try {
            try {
               ResultSetMetaData metadata = rs.getMetaData();
               this.columnNamesTypes = new ArrayList<Pair<String, Class<?>>>(metadata.getColumnCount());
               for(int colIdx=1, count=metadata.getColumnCount(); colIdx<=count; colIdx++) {
                  final Class<?> type;
                  switch (metadata.getColumnType(colIdx)) {
                  case Types.NUMERIC:   
                     type = (metadata.getScale(colIdx) > 0) ? Double.class : Long.class; 
                     break;
                  case Types.CHAR:
                  case Types.VARCHAR:   
                     type = String.class; 
                     break;
                  case Types.TIMESTAMP:
                  case Types.DATE:      
                     type = Date.class; 
                     break;
                  default:              
                     type = null;
                     break;
                  }
                  columnNamesTypes.add(new Pair<String, Class<?>>(metadata.getColumnName(colIdx), type));
               }

               this.dataRows = new ArrayList<List<Object>>();
               while (rs.next()) {
                  List<Object> line = new ArrayList<Object>();
                  for(int colIdx=1, size=columnNamesTypes.size(); colIdx<=size; colIdx++) {
                     Class<?> c = columnNamesTypes.get(colIdx-1).y;
                     if(c != null) {
                        if(c.equals(Long.class)) {
                           long l = rs.getLong(colIdx);
                           line.add(l == 0L ? (rs.wasNull() ? null : new Long(l)) : new Long(l));
                        }
                        if(c.equals(Date.class))
                           line.add(rs.getTimestamp(colIdx));
                        if(c.equals(String.class))
                           line.add(rs.getString(colIdx));
                        if(c.equals(Double.class)) {
                           double l = rs.getDouble(colIdx);
                           line.add(l == 0.0 ? (rs.wasNull() ? null : new Double(l)) : new Double(l));
                        }
                     } else 
                        line.add(rs.getString(colIdx));
                  }
                  dataRows.add(line);
               }
            } finally {
               rs.close();
            }
         } catch (SQLException e) {
            throw new RuntimeException(e);
         }
      else {
         this.columnNamesTypes = new ArrayList<Pair<String, Class<?>>>();
         this.dataRows = new ArrayList<List<Object>>();
      }
      this.currentRow = -1;
   }
   
   public boolean next() {
      if(dataRows == null)
         return false;
      return ++currentRow < dataRows.size();
   }
   
   public int size() {
      if(dataRows == null)
         return 0; 
      return dataRows.size();
   }
   
   public Object get(int columnIndex) {
      return dataRows.get(currentRow).get(columnIndex);
   }
   
   public Object get(String columnName) {
      return get(getColumnIndex(columnName));
   }
   
   public int getColumnIndex(String columnName) {
      return getColumnIndex(columnName, 0);
   }
   
   public int getColumnIndex(String columnName, int idx) {
      int c = 0;
      for(int i=0, s=columnNamesTypes.size(); i<s; i++)
         if(columnName.equalsIgnoreCase(columnNamesTypes.get(i).x)) {
            if(c == idx) return i;
            else         c++;
         }
      return -1;
   }
   
   public Map<String, Integer> getColumnToIndexMap() {
      Map<String, Integer> m = new HashMap<String, Integer>();
      for(int i=0, s=columnNamesTypes.size(); i<s; i++)
         m.put(columnNamesTypes.get(i).x, i);
      return m;
   }
   
   public Map<Integer, String> getIndexToColumnMap() {
      Map<Integer, String> m = new HashMap<Integer, String>();
      for(int i=0, s=columnNamesTypes.size(); i<s; i++)
         m.put(i, columnNamesTypes.get(i).x);
      return m;
   }
   
   public Map<String, Object> getMappedRow(int rowNum) {
      Map<String, Object> m = new HashMap<String, Object>();
      List<Object> row = dataRows.get(rowNum);
      for(int i=0, s=columnNamesTypes.size(); i<s; i++)
         m.put(columnNamesTypes.get(i).x, row.get(i));
      return m;
   }
   
   public Map<Integer, Object> getIndexedRow(int rowNum) {
      Map<Integer, Object> m = new HashMap<Integer, Object>();
      List<Object> row = dataRows.get(rowNum);
      for(int i=0, s=columnNamesTypes.size(); i<s; i++)
         m.put(i, row.get(i));
      return m;
   }
   
   public Map<String, Object> getMappedRow(List<Object> row) {
      Map<String, Object> m = new HashMap<String, Object>();
      for(int i=0, s=columnNamesTypes.size(); i<s; i++)
         m.put(columnNamesTypes.get(i).x, row.get(i));
      return m;
   }
   
   public Map<Integer, Object> getIndexedRow(List<Object> row) {
      Map<Integer, Object> m = new HashMap<Integer, Object>();
      for(int i=0, s=columnNamesTypes.size(); i<s; i++)
         m.put(i, row.get(i));
      return m;
   }
   
   public static ResultSetWrapper join(Join join, final ResultSetWrapper rs1, String[] rs1ColumnNames, final ResultSetWrapper rs2, String[] rs2ColumnNames) {
      int[] rs1ColumnIndices = new int[rs1ColumnNames.length];
      int[] rs2ColumnIndices = new int[rs2ColumnNames.length];
      
      for(int i=0, l=rs1ColumnNames.length; i<l; i++) 
         rs1ColumnIndices[i] = rs1.getColumnIndex(rs1ColumnNames[i]);
      for(int i=0, l=rs2ColumnNames.length; i<l; i++) 
         rs2ColumnIndices[i] = rs2.getColumnIndex(rs2ColumnNames[i]);
      
      return join(join, rs1, rs1ColumnIndices, rs2, rs2ColumnIndices);
   }
   
   public static ResultSetWrapper join(Join join, final ResultSetWrapper rs1, int[] rs1ColumnIndices, final ResultSetWrapper rs2, int[] rs2ColumnIndices) {
      return new ResultSetWrapper(
         ListBuilder.<Pair<String, Class<?>>>createArray().add(rs1.columnNamesTypes).add(rs2.columnNamesTypes).build(),
         CollectionUtil.generate(
            CollectionUtil.applyFilter(
               CollectionUtil.pairMaps(
                  CollectionUtil.split(rs1.dataRows, TransformerCollection.sublistTransformer(rs1ColumnIndices)),
                  CollectionUtil.split(rs2.dataRows, TransformerCollection.sublistTransformer(rs2ColumnIndices))
               ).values(), 
            join),
            new Generator<Pair<ArrayList<List<Object>>, ArrayList<List<Object>>>, List<Object>>() {
               @Override
               @SuppressWarnings("unchecked")
               public Iterable<List<Object>> generate(Pair<ArrayList<List<Object>>, ArrayList<List<Object>>> s) {
                  ArrayList<List<Object>> res = new ArrayList<List<Object>>();
                  if(s.x == null)
                     s.x = CollectionUtil.asList(Arrays.asList(new Object[rs1.columnNamesTypes.size()]));
                  if(s.y == null)
                     s.y = CollectionUtil.asList(Arrays.asList(new Object[rs2.columnNamesTypes.size()]));
                  for(List<Object> row1 : s.x)
                     for (List<Object> row2 : s.y)
                        res.add(CollectionUtil.addAll(new ArrayList<Object>(row1), row2));
                  return res;
               }
            }
         )
      );
   }

   @Override
   @SuppressWarnings("unchecked")
   public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
      int len = in.readInt();
      byte[] bytes = new byte[len];
      in.readFully(bytes);
      ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
      ObjectInputStream ois = new ObjectInputStream(new GZIPInputStream(bais));
      columnNamesTypes = (List<Pair<String, Class<?>>>) ois.readObject();
      dataRows = (List<List<Object>>) ois.readObject();
      ois.close();
   }

   @Override
   public void writeExternal(ObjectOutput out) throws IOException {
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      ObjectOutputStream oos = new ObjectOutputStream(new GZIPOutputStream(baos));
      oos.writeObject(columnNamesTypes);
      oos.writeObject(dataRows);
      oos.close();      
      byte[] bytes = baos.toByteArray();
      out.writeInt(bytes.length);
      out.write(bytes);
   }
   
   public class FieldTransformer <N> implements Transformer<List<Object>,N>, Serializable {
      private int idx;
      
      public FieldTransformer(String name) {
         this.idx = getColumnIndex(name);
      }

      public FieldTransformer(int idx) {
         this.idx = idx;
      }

      @Override
      @SuppressWarnings("unchecked")
      public N transform(List<Object> x) {
         return (N)x.get(idx);
      }
   }
   
   public class FieldListTransformer implements Transformer<List<Object>, List<Object>>, Serializable {
      private final int[] indices;
      
      public FieldListTransformer(int...indices) {
         this.indices = indices;
      }
      
      public FieldListTransformer(String ... columnNames) {
         indices = new int[columnNames.length];
         for(int i = 0; i < columnNames.length; i++)
            indices[i] = getColumnIndex(columnNames[i]);
      }
      
      @Override
      public List<Object> transform(List<Object> row) {
         List<Object> subRow = new ArrayList<Object>(indices.length);
         for(int index : indices)
            subRow.add(row.get(index));
         return subRow;
      }
   }
}

