package org.colllib.factories;

import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import org.colllib.datastruct.ImmutableArrayBackedList;

/**
 * Some standard factories.
 * <p>
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 *
 * @author marc.jackisch
 */
public class FactoryCollection {
   private FactoryCollection() {}

   private static final class CloneFactory<T> implements Factory<T>, Serializable {
      private final T baseObject;

      private CloneFactory(T baseObject) {
         this.baseObject = baseObject;
      }

      @SuppressWarnings("unchecked")
      @Override
      public T create() {
         Exception x = null;
         try {
            Class<?> c = baseObject.getClass();
            Method m = c.getDeclaredMethod("clone", new Class[0]);
            return (T) m.invoke(baseObject, new Object[0]);
         }
         catch(SecurityException e)         { x = e; }
         catch(IllegalArgumentException e)  { x = e; }
         catch(NoSuchMethodException e)     { x = e; }
         catch(IllegalAccessException e)    { x = e; }
         catch(InvocationTargetException e) { x = e; }
         throw new RuntimeException(x);
      }
   }

   /**
    * Return a factory that clones a given cloneable object.
    *
    * @param <T>
    * @param baseObject
    * @return the cloned object
    */
   public static <T extends Cloneable> Factory<T> cloneFactory(final T baseObject) {
      return new CloneFactory<T>(baseObject);
   }


   private static final class ConstructorCallFactory<T> implements Factory<T>, Serializable {
      private final Constructor<T> theConstructor;
      private final Object[]       constructorArguments;

      private ConstructorCallFactory(Constructor<T> theConstructor, Object[] constructorArguments) {
         this.theConstructor = theConstructor;
         this.constructorArguments = constructorArguments;
      }

      @Override
      public T create() {
         Exception x = null;
         try {
            return theConstructor.newInstance(constructorArguments);
         }
         catch(IllegalArgumentException e) { x = e; }
         catch(InstantiationException e) { x = e; }
         catch(IllegalAccessException e) { x = e; }
         catch(InvocationTargetException e) { x = e; }
         throw new RuntimeException(x);
      }
   }


   /**
    * Return a factory, which creates objects by calling a given constructor with
    * a fixed argument list.
    * <p>
    * Note, that the arguments constructorArgTypes and constructorArguments are passed
    * on to the underlying reflection functions Class.getConstructor and Class.newInstance
    * as they are passed to this method. No sanity checking is done here.
    *  
    * @param tClass                The class
    * @param constructorArgTypes   Argument type list used to identify the constructor
    * @param constructorArguments  Matching argument list to call the constructor
    * @return a factory
    */
   public static <T> Factory<T> constructorCallFactory(Class<T> tClass, final Class<?>[] constructorArgTypes, final Object[] constructorArguments) {
      try {
         final Constructor<T> theConstructor = tClass.getConstructor(constructorArgTypes);
         return new ConstructorCallFactory<T>(theConstructor, constructorArguments);
      }
      catch(SecurityException e) {
         throw new RuntimeException(e);
      }
      catch(NoSuchMethodException e) {
         throw new RuntimeException(e);
      }
   }


   /**
    * <b>THIS METHOD IS HIGHLY EXPERIMENTAL AND QUITE UNRELIABLE! WHEN IN DOUBT, USE
    * THE OTHER ONE!</b>
    * <p>
    * Return a factory, which creates objects by calling a given constructor with
    * a fixed argument list.
    * <p>
    * This variant tries to find the constructor by the list of arguments, that
    * have been passed to the method call. You'll have to ensure, that a matching
    * constructor exists and is unambiguos yourself.
    * <p>
    * Additionally note, that autoboxing / unboxing probably will not work in this
    * variant. This has the side-effect, that primitive datatypes will be autoboxed
    * when this method gets called, and the resulting class type will <i>not</i>
    * match a constructor, which is declared with the same primitive type.
    * 
    * @param tClass                The class
    * @param constructorArguments  argument list to call the constructor
    * @return a factory
    */
   public static <T> Factory<T> constructorCallFactory(Class<T> tClass, final Object ... constructorArguments) {
      @SuppressWarnings("unchecked")
      Constructor<T>[] cons = (Constructor<T>[]) tClass.getConstructors();
      List<Constructor<T>> matchedConstructors = new ArrayList<Constructor<T>>();
      for(Constructor<T> c : cons) {
         Class<?>[] parameterTypes = c.getParameterTypes();
         if(parameterTypes.length == constructorArguments.length) {
            boolean match = true;
            for(int i = 0; i < constructorArguments.length && match; i++)
               if(constructorArguments[i] != null && !parameterTypes[i].isAssignableFrom(constructorArguments[i].getClass()))
                  match = false;
            if(match)
               matchedConstructors.add(c);
         }
      }
      if(matchedConstructors.isEmpty())
         throw new IllegalArgumentException("No constructor found for class " + tClass + " and arguments " + new ImmutableArrayBackedList<Object>(constructorArguments));
      if(matchedConstructors.size() > 1)
         throw new IllegalArgumentException("Constructor call ambiguous: multiple constructors found for class " + tClass + " and arguments " + new ImmutableArrayBackedList<Object>(constructorArguments));

      final Constructor<T> theConstructor = matchedConstructors.get(0);
      return new ConstructorCallFactory<T>(theConstructor, constructorArguments);
   }


   private static final class ConstFactory<T> implements Factory<T>, Serializable {
      private final T value;

      private ConstFactory(T value) {
         this.value = value;
      }

      @Override
      public T create() {
         return value;
      }
   }

   
   /**
    * Return a factory, which always returns the same object. (Note that this really
    * is exactly the same object pointer)
    * 
    * @param value the object to return
    * @return a factory
    */
   public static <T> Factory<T> constFactory(final T value) {
      return new ConstFactory<T>(value);
   }
}
