package org.colllib.util;

import java.util.Comparator;

/**
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 * 
 * @author marc.jackisch
 */
public class CompareUtil {
   private CompareUtil() {}
   
   /**
    * Check equality of two objects via the underlying equals method, but
    * honoring null pointers.
    * 
    * @param o1
    * @param o2
    * @return true, if both are null, false if only one object is null,
    * <code>o1.equals(o2)</code> otherwise
    */
   public static boolean nullSafeEquals(Object o1, Object o2) {
      if(o1 == null && o2 == null)
         return true;
      if(o1 == null || o2 == null)
         return false;
      return o1.equals(o2);
   }
   
   /**
    * Compares two objects using their own {@code compareTo} method, but 
    * checking for null pointers first.
    * 
    * @param t1
    * @param t2
    * @return a negative integer, zero, or a positive integer as this object 
    * is less than, equal to, or greater than the specified object. If one object is
    * null it is considered as less, if both are null they are considered as equal
    */
   public static <T extends Comparable<T>> int nullSafeCompareTo(T t1, T t2) {
      if(t1 == null && t2 == null)
         return 0;
      if(t1 == null || t2 == null)
         return t1 == null ? -1 : 1;
      return t1.compareTo(t2);
   }
   
   /**
    * Returns a {@link Comparator} which compares objects while honoring null pointers.
    * 
    * @return a null safe comparator
    */
   public static <T extends Comparable<T>> Comparator<T> nullSafeComparator() {
      return new Comparator<T>() {
         @Override
         public int compare(T o1, T o2) {
            return nullSafeCompareTo(o1, o2);
         }};
   }
   
   /**
    * Compares a list of objects by comparing every object in the list with the one at 
    * index + 1 using the {@link CompareUtil.nullSafeCompareTo} method.
    * 
    * @param ts
    * @return the result of {@link CompareUtil.nullSafeCompareTo} for the first object
    * pair which is not equal or zero if all objects pairs are equal
    */
   public static <T extends Comparable<T>> int pairCompare(T... ts) {
      if(ts.length % 2 != 0)
         throw new IllegalArgumentException("Argument count must be a multiple of 2");
      int c = 0;
      for(int i=0; c==0 && i<ts.length; i+=2)
         c = nullSafeCompareTo(ts[i], ts[i+1]);
      return c;
   }
}

