package org.colllib.util;

import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Pattern;

import org.colllib.filter.FilterCollection;

/**
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 * 
 * @author marc.jackisch
 */
public class StringUtil {
   private StringUtil() {}

   /**
    * Returns the given string or an empty string if the parameter was null.
    * @param s  the string
    * @return the given string or an empty string
    */
   public static String emptyStringIfNull(String s) {
      return s == null ? "" : s;
   }

   /**
    * Returns the given string or null if the parameter was empty.
    * @param s  the string
    * @return the given string or null
    */
   public static String nullIfEmpty(String s) {
      if(s == null || s.length() == 0)
         return null;
      return s;
   }

   /**
    * Returns the given string or the given default if the parameter was null.
    * @param s  the string
    * @param def  the default
    * @return the given string or the default string
    */
   public static String defaultIfNull(String s, String def) {
      return s != null ? s : def;
   }

   /**
    * Returns the given string or the given default if the parameter was null or empty.
    * @param s  the string
    * @param def  the default
    * @return the given string or the default string
    */
   public static String defaultIfNullOrEmpty(String s, String def) {
      return defaultIfNull(nullIfEmpty(s), def);
   }

   /**
    * Converts a collection of T into a single string containing the <code>toString()</code> representation
    * of each object delimited by the given delimiter.
    * @param coll  an iterable of T
    * @param delim  the delimiter
    * @return the string containing all T
    */
   public static <T> String asString(Iterable<T> coll, String delim) {
      if(coll == null) 
         return "";
      StringBuilder sb = new StringBuilder();
      Iterator<T> i = coll.iterator();
      while(i.hasNext()) {
         T next = i.next();
         sb.append(next != null ? next.toString() : "null");
         if(i.hasNext()) sb.append(delim);
      }
      return sb.toString();
   }

   /**
    * Splits a string by a delimiter.
    * @param s  the string
    * @param d  the delimiter
    * @return  a list containing all substrings
    */
   public static List<String> toCollection(String s, String d) {
      return s == null ? Collections.<String>emptyList() : CollectionUtil.asList(s.split(Pattern.quote(d)));
   }

   /**
    * Remove all empty strings from a given list of strings
    * @param strings  a list of strings
    * @return  an iterable containing all non-empty strings 
    */
   public static Iterable<String> nonEmptyIterable(String... strings) {
      return CollectionUtil.applyFilter(strings, FilterCollection.not(FilterCollection.<String>matches("")));
   }

   /**
    * Call the <code>toString()</code> of an object and return the result or an empty string 
    * if the given object was null.
    * @param obj  the object
    * @return  the <code>toString()</code> representation of the object or an empty string
    */
   public static String safeToString(Object obj) {
      return obj == null ? "" : obj.toString();
   }
}

