package org.colllib.util;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;

import org.colllib.datastruct.Pair;


/**
 * TypeUtil: utility methods to work with java type conversions and generics.
 * Keeps those nasty unchecked warnings in one place
 * 
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 * 
 * @author marc.jackisch
 */
public class TypeUtil {
   private TypeUtil() {}

   /**
    * Type conversion / check of a list. Pass an untyped or wrongly typed list,
    * the elements target class, and get a correctly typed list or a class cast
    * exception.
    * 
    * @param in       the original list
    * @param tClass   the desired element class
    * @return a cast transformation of 
    */
   @SuppressWarnings("unchecked")
   public static <T> List<T> typecheck(List<?> in, Class<T> tClass) {
      if(in != null)
         for(Object t : in)
            tClass.cast(t); // throws class cast exception for wrong types
      return (List<T>) in;
   }

   /**
    * Convert an object of type vClass to type tClass
    * 
    * @param v        the object to convert
    * @param vClass   source class
    * @param tClass   target class
    * @return v  the converted object of type tClass
    * @exception java.lang.UnsupportedOperationException in case of any error during conversion
    */
   @SuppressWarnings("unchecked")
   public static <T> T convert(Object v, Class<?> vClass, Class<T> tClass) {
      if(v == null)
         return (T) v;
      if(vClass.equals(tClass))
         return (T) v;
      if(tClass.isAssignableFrom(vClass))
         return tClass.cast(v);

      if(tClass.isPrimitive()) {
         return resolvePrimitive(v, vClass, tClass);
      }

      try {
         for(Constructor<?> constr : tClass.getConstructors())
            if(constr.getParameterTypes().length == 1
            && constr.getParameterTypes()[0].isAssignableFrom(vClass))
               return (T) constr.newInstance(new Object[] {v});
      }
      catch(IllegalArgumentException e) {}
      catch(InstantiationException e) {}
      catch(IllegalAccessException e) {}
      catch(InvocationTargetException e) {}

      if(tClass.equals(String.class))
         return (T) v.toString();

      if(Number.class.isAssignableFrom(vClass) && Number.class.isAssignableFrom(tClass)) {
         try {
            String tClassName = tClass.equals(Integer.class)
               ? "int" : tClass.getSimpleName().toLowerCase();
            Method extract = Number.class.getMethod(tClassName + "Value", new Class[0]);
            Field typeField = tClass.getField("TYPE");
            Class<?> typeClassConstant = (Class<?>) typeField.get(null);
            Constructor<?> tcc = tClass.getConstructor(new Class[] { typeClassConstant });
            Object res = tcc.newInstance(new Object[] { extract.invoke(v, new Object[0]) });
            return (T) res;
         }
         catch(SecurityException e) {}
         catch(NoSuchMethodException e) {}
         catch(NoSuchFieldException e) {}
         catch(IllegalArgumentException e) {}
         catch(IllegalAccessException e) {}
         catch(InstantiationException e) {}
         catch(InvocationTargetException e) {}
      }

      if(tClass.isEnum()) {
         String compareValue = v.toString();
         for(T t : tClass.getEnumConstants())
            if(((Enum<?>) t).name().equals(compareValue))
               return t;
      }

      if(Number.class.isAssignableFrom(tClass) && vClass.equals(String.class) && StringUtil.nullIfEmpty((String)v) == null)
         return null;

      throw new UnsupportedOperationException("Cannot convert type " + vClass + " to " + tClass);
   }

   /**
    * Convert an object to type tClass
    * 
    * @param v        the object to convert
    * @param tClass   target class
    * @return v  the converted object of type tClass
    * @exception java.lang.UnsupportedOperationException in case of any error during conversion
    */
   public static <T> T convert(Object v, Class<T> tClass) {
      if(v == null)
         return null;
      return convert(v, v.getClass(), tClass);
   }

   @SuppressWarnings("unchecked")
   private static <T> T resolvePrimitive(Object v, Class<?> vClass, Class<T> tClass) {
      try {
         for(Pair<Class<?>, String> p : new Pair[] {
            new Pair<Class<?>, String>(Boolean.class, "booleanValue"),
            new Pair<Class<?>, String>(Byte.class, "byteValue"),
            new Pair<Class<?>, String>(Character.class, "charValue"),
            new Pair<Class<?>, String>(Double.class, "doubleValue"),
            new Pair<Class<?>, String>(Float.class, "floatValue"),
            new Pair<Class<?>, String>(Integer.class, "intValue"),
            new Pair<Class<?>, String>(Long.class, "longValue"),
            new Pair<Class<?>, String>(Short.class, "shortValue")
         }) {
            Field typeField = p.x.getDeclaredField("TYPE");
            Class<?> typePrimitiveClass = (Class<?>) typeField.get(null);
            if(typePrimitiveClass.equals(tClass)) {
               Object baseConv = convert(v, vClass, p.x);
               return (T) p.x.getMethod(p.y, new Class[0]).invoke(baseConv, new Object[0]);
            }
         }
      }
      catch(Exception e) {}
      throw new UnsupportedOperationException("Cannot convert primitive type " + vClass + " to " + tClass);
   }
}
